import asyncio
import logging
from typing import Union

from cognee.modules.cognify.config import get_cognify_config
from cognee.infrastructure.databases.relational import get_relational_engine
from cognee.shared.data_models import KnowledgeGraph
from cognee.modules.data.models import Dataset, Data
from cognee.modules.data.operations.get_dataset_data import get_dataset_data
from cognee.modules.data.operations.retrieve_datasets import retrieve_datasets
from cognee.modules.pipelines.tasks.Task import Task
from cognee.modules.pipelines import run_tasks, run_tasks_parallel
from cognee.modules.users.models import User
from cognee.modules.users.methods import get_default_user
from cognee.modules.pipelines.operations.get_pipeline_status import get_pipeline_status
from cognee.modules.pipelines.operations.log_pipeline_status import log_pipeline_status
from cognee.tasks import chunk_extract_summary, \
    chunk_naive_llm_classifier, \
    chunk_remove_disconnected, \
    infer_data_ontology, \
    save_chunks_to_store, \
    chunk_update_check, \
    chunks_into_graph, \
    source_documents_to_chunks, \
    check_permissions_on_documents, \
    classify_documents

logger = logging.getLogger("cognify.v2")

update_status_lock = asyncio.Lock()

class PermissionDeniedException(Exception):
    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

async def cognify(datasets: Union[str, list[str]] = None, user: User = None):
    db_engine = get_relational_engine()

    if datasets is None or len(datasets) == 0:
        return await cognify(await db_engine.get_datasets())

    if type(datasets[0]) == str:
        datasets = await retrieve_datasets(datasets)

    if user is None:
        user = await get_default_user()

    async def run_cognify_pipeline(dataset: Dataset):
        data_documents: list[Data] = await get_dataset_data(dataset_id = dataset.id)

        document_ids_str = [str(document.id) for document in data_documents]

        dataset_id = dataset.id
        dataset_name = generate_dataset_name(dataset.name)

        async with update_status_lock:
            task_status = await get_pipeline_status([dataset_id])

            if dataset_id in task_status and task_status[dataset_id] == "DATASET_PROCESSING_STARTED":
                logger.info("Dataset %s is already being processed.", dataset_name)
                return

            await log_pipeline_status(dataset_id, "DATASET_PROCESSING_STARTED", {
                "dataset_name": dataset_name,
                "files": document_ids_str,
            })
        try:
            cognee_config = get_cognify_config()

            root_node_id = None

            tasks = [
                Task(classify_documents),
                Task(check_permissions_on_documents, user = user, permissions = ["write"]),
                Task(infer_data_ontology, root_node_id = root_node_id, ontology_model = KnowledgeGraph),
                Task(source_documents_to_chunks, parent_node_id = root_node_id), # Classify documents and save them as a nodes in graph db, extract text chunks based on the document type
                Task(chunks_into_graph, graph_model = KnowledgeGraph, collection_name = "entities", task_config = { "batch_size": 10 }), # Generate knowledge graphs from the document chunks and attach it to chunk nodes
                Task(chunk_update_check, collection_name = "chunks"), # Find all affected chunks, so we don't process unchanged chunks
                Task(
                    save_chunks_to_store,
                    collection_name = "chunks",
                ), # Save the document chunks in vector db and as nodes in graph db (connected to the document node and between each other)
                run_tasks_parallel([
                    Task(
                        chunk_extract_summary,
                        summarization_model = cognee_config.summarization_model,
                        collection_name = "chunk_summaries",
                    ), # Summarize the document chunks
                    Task(
                        chunk_naive_llm_classifier,
                        classification_model = cognee_config.classification_model,
                    ),
                ]),
                Task(chunk_remove_disconnected), # Remove the obsolete document chunks.
            ]

            pipeline = run_tasks(tasks, data_documents)

            async for result in pipeline:
                print(result)

            await log_pipeline_status(dataset_id, "DATASET_PROCESSING_FINISHED", {
                "dataset_name": dataset_name,
                "files": document_ids_str,
            })
        except Exception as error:
            await log_pipeline_status(dataset_id, "DATASET_PROCESSING_ERROR", {
                "dataset_name": dataset_name,
                "files": document_ids_str,
            })
            raise error


    existing_datasets = [dataset.name for dataset in list(await db_engine.get_datasets())]
    awaitables = []

    for dataset in datasets:
        dataset_name = generate_dataset_name(dataset.name)

        if dataset_name in existing_datasets:
            awaitables.append(run_cognify_pipeline(dataset))

    return await asyncio.gather(*awaitables)


def generate_dataset_name(dataset_name: str) -> str:
    return dataset_name.replace(".", "_").replace(" ", "_")



from sklearn.feature_extraction.text import TfidfVectorizer
from cognee.shared.utils import extract_pos_tags

def extract_keywords(text: str) -> list[str]:
    if len(text) == 0:
        raise ValueError("extract_keywords cannot extract keywords from empty text.")

    tags = extract_pos_tags(text)
    nouns = [word for (word, tag) in tags if tag == "NN"]

    vectorizer = TfidfVectorizer()
    tfidf = vectorizer.fit_transform(nouns)

    top_nouns = sorted(
        vectorizer.vocabulary_,
        key = lambda x: tfidf[0, vectorizer.vocabulary_[x]],
        reverse = True
    )

    keywords = []

    for word in top_nouns:
        if len(word) > 3:
            keywords.append(word)
        if len(keywords) >= 15:
            break

    return keywords



from .api.v1.config.config import config
from .api.v1.add.add import add
from .api.v1.cognify.cognify_v2 import cognify
from .api.v1.datasets.datasets import datasets
from .api.v1.search.search import search, SearchType
from .api.v1.prune import prune

# Pipelines
from .modules import pipelines



""" FastAPI server for the Cognee API. """
import os
import aiohttp
import uvicorn
import json
import logging
import sentry_sdk
from typing import Dict, Any, List, Union, Optional, Literal
from typing_extensions import Annotated
from fastapi import FastAPI, HTTPException, Form, UploadFile, Query
from fastapi.responses import JSONResponse, FileResponse
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel

from cognee.infrastructure.databases.relational import create_db_and_tables

# Set up logging
logging.basicConfig(
    level=logging.INFO,  # Set the logging level (e.g., DEBUG, INFO, WARNING, ERROR, CRITICAL)
    format="%(asctime)s [%(levelname)s] %(message)s",  # Set the log message format
)
logger = logging.getLogger(__name__)

if os.getenv("ENV") == "prod":
    sentry_sdk.init(
        dsn = os.getenv("SENTRY_REPORTING_URL"),
        traces_sample_rate = 1.0,
        profiles_sample_rate = 1.0,
    )

from contextlib import asynccontextmanager

@asynccontextmanager
async def lifespan(app: FastAPI):
    # Not needed if you setup a migration system like Alembic
    await create_db_and_tables()
    yield

app = FastAPI(debug = os.getenv("ENV") != "prod", lifespan = lifespan)

origins = [
    "http://127.0.0.1:3000",
    "http://frontend:3000",
    "http://localhost:3000",
    "http://localhost:3001",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["OPTIONS", "GET", "POST", "DELETE"],
    allow_headers=["*"],
)

from cognee.api.v1.users.routers import get_auth_router, get_register_router,\
    get_reset_password_router, get_verify_router, get_users_router

from cognee.api.v1.permissions.get_permissions_router import get_permissions_router

app.include_router(
    get_auth_router(),
    prefix = "/auth/jwt",
    tags = ["auth"]
)

app.include_router(
    get_register_router(),
    prefix = "/auth",
    tags = ["auth"],
)

app.include_router(
    get_reset_password_router(),
    prefix = "/auth",
    tags = ["auth"],
)

app.include_router(
    get_verify_router(),
    prefix = "/auth",
    tags = ["auth"],
)

app.include_router(
    get_users_router(),
    prefix = "/users",
    tags = ["users"],
)

app.include_router(
    get_permissions_router(),
    prefix = "/permissions",
    tags = ["permissions"],
)

@app.get("/")
async def root():
    """
    Root endpoint that returns a welcome message.
    """
    return {"message": "Hello, World, I am alive!"}

@app.get("/health")
def health_check():
    """
    Health check endpoint that returns the server status.
    """
    return {"status": "OK"}

@app.get("/datasets", response_model = list)
async def get_datasets():
    try:
        from cognee.api.v1.datasets.datasets import datasets
        datasets = await datasets.list_datasets()

        return JSONResponse(
            status_code = 200,
            content = [dataset.to_json() for dataset in datasets],
        )
    except Exception as error:
        raise HTTPException(status_code = 500, detail=f"Error retrieving datasets: {str(error)}") from error

@app.delete("/datasets/{dataset_id}", response_model = dict)
async def delete_dataset(dataset_id: str):
    from cognee.api.v1.datasets.datasets import datasets
    await datasets.delete_dataset(dataset_id)

    return JSONResponse(
        status_code = 200,
        content = "OK",
    )

@app.get("/datasets/{dataset_id}/graph", response_model=list)
async def get_dataset_graph(dataset_id: str):
    from cognee.shared.utils import render_graph
    from cognee.infrastructure.databases.graph import get_graph_engine

    try:
        graph_client = await get_graph_engine()
        graph_url = await render_graph(graph_client.graph)

        return JSONResponse(
            status_code = 200,
            content = str(graph_url),
        )
    except:
        return JSONResponse(
            status_code = 409,
            content = "Graphistry credentials are not set. Please set them in your .env file.",
        )

@app.get("/datasets/{dataset_id}/data", response_model=list)
async def get_dataset_data(dataset_id: str):
    from cognee.api.v1.datasets.datasets import datasets

    dataset_data = await datasets.list_data(dataset_id = dataset_id)

    if dataset_data is None:
        raise HTTPException(status_code = 404, detail = f"Dataset ({dataset_id}) not found.")

    return [
        data.to_json() for data in dataset_data
    ]

@app.get("/datasets/status", response_model=dict)
async def get_dataset_status(datasets: Annotated[List[str], Query(alias="dataset")] = None):
    from cognee.api.v1.datasets.datasets import datasets as cognee_datasets

    try:
        datasets_statuses = await cognee_datasets.get_status(datasets)

        return JSONResponse(
            status_code = 200,
            content = datasets_statuses,
        )
    except Exception as error:
        return JSONResponse(
            status_code = 409,
            content = {"error": str(error)}
        )

@app.get("/datasets/{dataset_id}/data/{data_id}/raw", response_class=FileResponse)
async def get_raw_data(dataset_id: str, data_id: str):
    from cognee.api.v1.datasets.datasets import datasets
    dataset_data = await datasets.list_data(dataset_id)

    if dataset_data is None:
        raise HTTPException(status_code = 404, detail = f"Dataset ({dataset_id}) not found.")

    data = [data for data in dataset_data if str(data.id) == data_id][0]
    return data.raw_data_location

class AddPayload(BaseModel):
    data: Union[str, UploadFile, List[Union[str, UploadFile]]]
    dataset_id: str
    class Config:
        arbitrary_types_allowed = True

@app.post("/add", response_model=dict)
async def add(
    data: List[UploadFile],
    datasetId: str = Form(...),
):
    """ This endpoint is responsible for adding data to the graph."""
    from cognee.api.v1.add import add as cognee_add
    try:
        if isinstance(data, str) and data.startswith("http"):
            if "github" in data:
                # Perform git clone if the URL is from GitHub
                repo_name = data.split("/")[-1].replace(".git", "")
                os.system(f"git clone {data} .data/{repo_name}")
                await cognee_add(
                    "data://.data/",
                    f"{repo_name}",
                )
            else:
                # Fetch and store the data from other types of URL using curl
                async with aiohttp.ClientSession() as session:
                    async with session.get(data) as resp:
                        if resp.status == 200:
                            file_data = await resp.read()
                            with open(f".data/{data.split('/')[-1]}", "wb") as f:
                                f.write(file_data)
                            await cognee_add(
                                "data://.data/",
                                f"{data.split('/')[-1]}",
                            )
        else:
            await cognee_add(
                data,
                datasetId,
            )
            return JSONResponse(
                status_code=200,
                content="OK"
            )
    except Exception as error:
        return JSONResponse(
            status_code=409,
            content={"error": str(error)}
        )

class CognifyPayload(BaseModel):
    datasets: List[str]

@app.post("/cognify", response_model=dict)
async def cognify(payload: CognifyPayload):
    """ This endpoint is responsible for the cognitive processing of the content."""
    from cognee.api.v1.cognify.cognify_v2 import cognify as cognee_cognify
    try:
        await cognee_cognify(payload.datasets)
        return JSONResponse(
            status_code = 200,
            content = "OK"
        )
    except Exception as error:
        return JSONResponse(
            status_code = 409,
            content = {"error": str(error)}
        )

class SearchPayload(BaseModel):
    query_params: Dict[str, Any]

@app.post("/search", response_model=dict)
async def search(payload: SearchPayload):
    """ This endpoint is responsible for searching for nodes in the graph."""
    from cognee.api.v1.search import search as cognee_search
    try:
        search_type = payload.query_params["searchType"]

        params = {
            "query": payload.query_params["query"],
        }

        results = await cognee_search(search_type, params)

        return JSONResponse(
            status_code = 200,
            content = results,
        )
    except Exception as error:
        return JSONResponse(
            status_code = 409,
            content = {"error": str(error)}
        )

@app.get("/settings", response_model=dict)
async def get_settings():
    from cognee.modules.settings import get_settings as get_cognee_settings
    return get_cognee_settings()

class LLMConfig(BaseModel):
    provider: Union[Literal["openai"], Literal["ollama"], Literal["anthropic"]]
    model: str
    apiKey: str

class VectorDBConfig(BaseModel):
    provider: Union[Literal["lancedb"], Literal["qdrant"], Literal["weaviate"]]
    url: str
    apiKey: str

class SettingsPayload(BaseModel):
    llm: Optional[LLMConfig] = None
    vectorDB: Optional[VectorDBConfig] = None

@app.post("/settings", response_model=dict)
async def save_config(new_settings: SettingsPayload):
    from cognee.modules.settings import save_llm_config, save_vector_db_config
    if new_settings.llm is not None:
        await save_llm_config(new_settings.llm)
    if new_settings.vectorDB is not None:
        await save_vector_db_config(new_settings.vectorDB)
    return JSONResponse(
        status_code=200,
        content="OK",
    )


def start_api_server(host: str = "0.0.0.0", port: int = 8000):
    """
    Start the API server using uvicorn.
    Parameters:
    host (str): The host for the server.
    port (int): The port for the server.
    """
    try:
        logger.info("Starting server at %s:%s", host, port)

        import asyncio
        from cognee.modules.data.deletion import prune_system, prune_data
        asyncio.run(prune_data())
        asyncio.run(prune_system(metadata = True))

        uvicorn.run(app, host = host, port = port)
    except Exception as e:
        logger.exception(f"Failed to start server: {e}")
        # Here you could add any cleanup code or error recovery code.


if __name__ == "__main__":
    start_api_server()



import inspect
from typing import Any

from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session

class FakeAsyncSession:
    def __init__(self, session: Session):
        self.session = session

    def run_sync(self, *args, **kwargs):
        return self.execute(*args, **kwargs)

    def __getattr__(self, name: str) -> Any:
        """
        If the method being called is async in AsyncSession, create a fake async version
        for Session so callers can `await` as usual. Think `commit`, `refresh`,
        `delete`, etc.
        """
        async_session_attr = getattr(AsyncSession, name, None)
        session_attr = getattr(self.session, name)

        if not inspect.iscoroutinefunction(async_session_attr):
            return session_attr

        async def async_wrapper(*args, **kwargs):
            return session_attr(*args, **kwargs)

        return async_wrapper



from .check_permission_on_documents import check_permission_on_documents
from .give_permission_on_document import give_permission_on_document
from .get_document_ids_for_user import get_document_ids_for_user



from cognee.infrastructure.databases.relational.user_authentication.users import authenticate_user_method


async def authenticate_user(email: str, password: str):
    """
    This function is used to authenticate a user.
    """
    output = await authenticate_user_method(email=email, password=password)
    return output


if __name__ == "__main__":
    import asyncio
    # Define an example user
    example_email = "example@example.com"
    example_password = "securepassword123"
    example_is_superuser = False

    # Create an event loop and run the create_user function
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(authenticate_user(example_email, example_password))

    # Print the result
    print(result)



from cognee.infrastructure.databases.vector import get_vector_engine
from cognee.infrastructure.databases.graph.get_graph_engine import get_graph_engine
from cognee.infrastructure.databases.relational import get_relational_engine

async def prune_system(graph = True, vector = True, metadata = False):
    if graph:
        graph_engine = await get_graph_engine()
        await graph_engine.delete_graph()

    if vector:
        vector_engine = get_vector_engine()
        await vector_engine.prune()

    if metadata:
        db_engine = get_relational_engine()
        await db_engine.delete_database()



from uuid import uuid5, NAMESPACE_OID
from .data_types import IngestionData

def identify(data: IngestionData) -> str:
    data_id: str = data.get_identifier()

    return uuid5(NAMESPACE_OID, data_id)



from typing import List, Optional, get_type_hints, Generic, TypeVar
import asyncio
import lancedb
from lancedb.pydantic import Vector, LanceModel
from cognee.infrastructure.files.storage import LocalStorage
from ..models.ScoredResult import ScoredResult
from ..vector_db_interface import VectorDBInterface, DataPoint
from ..embeddings.EmbeddingEngine import EmbeddingEngine

class LanceDBAdapter(VectorDBInterface):
    name = "LanceDB"
    url: str
    api_key: str
    connection: lancedb.AsyncConnection = None


    def __init__(
        self,
        url: Optional[str],
        api_key: Optional[str],
        embedding_engine: EmbeddingEngine,
    ):
        self.url = url
        self.api_key = api_key
        self.embedding_engine = embedding_engine

    async def get_connection(self):
        if self.connection is None:
            self.connection = await lancedb.connect_async(self.url, api_key = self.api_key)

        return self.connection

    async def embed_data(self, data: list[str]) -> list[list[float]]:
        return await self.embedding_engine.embed_text(data)

    async def has_collection(self, collection_name: str) -> bool:
        connection = await self.get_connection()
        collection_names = await connection.table_names()
        return collection_name in collection_names

    async def create_collection(self, collection_name: str, payload_schema = None):
        data_point_types = get_type_hints(DataPoint)
        vector_size = self.embedding_engine.get_vector_size()

        class LanceDataPoint(LanceModel):
            id: data_point_types["id"]
            vector: Vector(vector_size)
            payload: payload_schema

        if not await self.has_collection(collection_name):
            connection = await self.get_connection()
            return await connection.create_table(
                name = collection_name,
                schema = LanceDataPoint,
                exist_ok = True,
            )

    async def create_data_points(self, collection_name: str, data_points: List[DataPoint]):
        connection = await self.get_connection()

        if not await self.has_collection(collection_name):
            await self.create_collection(
                collection_name,
                payload_schema = type(data_points[0].payload),
            )

        collection = await connection.open_table(collection_name)

        data_vectors = await self.embed_data(
            [data_point.get_embeddable_data() for data_point in data_points]
        )

        IdType = TypeVar("IdType")
        PayloadSchema = TypeVar("PayloadSchema")
        vector_size = self.embedding_engine.get_vector_size()

        class LanceDataPoint(LanceModel, Generic[IdType, PayloadSchema]):
            id: IdType
            vector: Vector(vector_size)
            payload: PayloadSchema

        lance_data_points = [
            LanceDataPoint[type(data_point.id), type(data_point.payload)](
                id = data_point.id,
                vector = data_vectors[data_index],
                payload = data_point.payload,
            ) for (data_index, data_point) in enumerate(data_points)
        ]

        await collection.add(lance_data_points)

    async def retrieve(self, collection_name: str, data_point_ids: list[str]):
        connection = await self.get_connection()
        collection = await connection.open_table(collection_name)

        if len(data_point_ids) == 1:
            results = await collection.query().where(f"id = '{data_point_ids[0]}'").to_pandas()
        else:
            results = await collection.query().where(f"id IN {tuple(data_point_ids)}").to_pandas()

        return [ScoredResult(
            id = result["id"],
            payload = result["payload"],
            score = 0,
        ) for result in results.to_dict("index").values()]

    async def search(
        self,
        collection_name: str,
        query_text: str = None,
        query_vector: List[float] = None,
        limit: int = 5,
        with_vector: bool = False,
    ):
        if query_text is None and query_vector is None:
            raise ValueError("One of query_text or query_vector must be provided!")

        if query_text and not query_vector:
            query_vector = (await self.embedding_engine.embed_text([query_text]))[0]

        connection = await self.get_connection()
        collection = await connection.open_table(collection_name)

        results = await collection.vector_search(query_vector).limit(limit).to_pandas()

        result_values = list(results.to_dict("index").values())

        min_value = 100
        max_value = 0

        for result in result_values:
            value = float(result["_distance"])
            if value > max_value:
                max_value = value
            if value < min_value:
                min_value = value

        normalized_values = [(result["_distance"] - min_value) / (max_value - min_value) for result in result_values]

        return [ScoredResult(
            id = str(result["id"]),
            payload = result["payload"],
            score = normalized_values[value_index],
        ) for value_index, result in enumerate(result_values)]

    async def batch_search(
        self,
        collection_name: str,
        query_texts: List[str],
        limit: int = None,
        with_vectors: bool = False,
    ):
        query_vectors = await self.embedding_engine.embed_text(query_texts)

        return asyncio.gather(
            *[self.search(
                collection_name = collection_name,
                query_vector = query_vector,
                limit = limit,
                with_vector = with_vectors,
            ) for query_vector in query_vectors]
        )

    async def delete_data_points(self, collection_name: str, data_point_ids: list[str]):
        connection = await self.get_connection()
        collection = await connection.open_table(collection_name)
        results = await collection.delete(f"id IN {tuple(data_point_ids)}")
        return results

    async def prune(self):
        # Clean up the database if it was set up as temporary
        if self.url.startswith("/"):
            LocalStorage.remove_all(self.url) # Remove the temporary directory and files inside



from typing import BinaryIO
from cognee.infrastructure.files import get_file_metadata, FileMetadata
from .IngestionData import IngestionData

def create_binary_data(data: BinaryIO):
    return BinaryData(data)

class BinaryData(IngestionData):
    name: str = None
    data: BinaryIO = None
    metadata: FileMetadata = None

    def __init__(self, data: BinaryIO, name: str = None):
        self.name = name
        self.data = data

    def get_identifier(self):
        metadata = self.get_metadata()

        return self.name + "." + metadata["extension"]

    def get_metadata(self):
        self.ensure_metadata()

        return self.metadata

    def ensure_metadata(self):
        if self.metadata is None:
            self.metadata = get_file_metadata(self.data)

            if self.metadata["name"] is None:
                self.metadata["name"] = self.name

    def get_data(self):
        return self.data



from typing import Protocol, Optional, Dict, Any
from abc import abstractmethod

class GraphDBInterface(Protocol):
    @abstractmethod
    async def graph(self):
        raise NotImplementedError

    @abstractmethod
    async def add_node(
        self,
        node_id: str,
        node_properties: dict
    ): raise NotImplementedError

    @abstractmethod
    async def add_nodes(
        self,
        nodes: list[tuple[str, dict]]
    ): raise NotImplementedError

    @abstractmethod
    async def delete_node(
        self,
        node_id: str
    ): raise NotImplementedError

    @abstractmethod
    async def delete_nodes(
        self,
        node_ids: list[str]
    ): raise NotImplementedError

    @abstractmethod
    async def extract_node(
        self,
        node_id: str
    ): raise NotImplementedError

    @abstractmethod
    async def extract_nodes(
        self,
        node_ids: list[str]
    ): raise NotImplementedError

    @abstractmethod
    async def add_edge(
        self,
        from_node: str,
        to_node: str,
        relationship_name: str,
        edge_properties: Optional[Dict[str, Any]] = None
    ): raise NotImplementedError

    @abstractmethod
    async def add_edges(
        self,
        edges: tuple[str, str, str, dict]
    ): raise NotImplementedError

    @abstractmethod
    async def delete_graph(
        self,
    ): raise NotImplementedError



"""Factory function to get the appropriate graph client based on the graph type."""

from .config import get_graph_config
from .graph_db_interface import GraphDBInterface
from .networkx.adapter import NetworkXAdapter


async def get_graph_engine() -> GraphDBInterface :
    """Factory function to get the appropriate graph client based on the graph type."""
    config = get_graph_config()

    if config.graph_database_provider == "neo4j":
        try:
            from .neo4j_driver.adapter import Neo4jAdapter

            return Neo4jAdapter(
                graph_database_url = config.graph_database_url,
                graph_database_username = config.graph_database_username,
                graph_database_password = config.graph_database_password
            )
        except:
            pass

    elif config.graph_database_provider == "falkorb":
        try:
            from .falkordb.adapter import FalcorDBAdapter

            return FalcorDBAdapter(
                graph_database_url = config.graph_database_url,
                graph_database_username = config.graph_database_username,
                graph_database_password = config.graph_database_password,
                graph_database_port = config.graph_database_port
            )
        except:
            pass

    graph_client = NetworkXAdapter(filename = config.graph_file_path)

    if graph_client.graph is None:
        await graph_client.load_graph_from_file()

    return graph_client



from uuid import uuid4
from typing import List
from datetime import datetime, timezone
from sqlalchemy.orm import relationship, Mapped
from sqlalchemy import Column, String, DateTime, UUID
from cognee.infrastructure.databases.relational import Base
from .DatasetData import DatasetData

class Data(Base):
    __tablename__ = "data"

    id = Column(UUID(as_uuid = True), primary_key = True, default = uuid4)

    name = Column(String)
    extension = Column(String)
    mime_type = Column(String)
    raw_data_location = Column(String)

    created_at = Column(DateTime(timezone = True), default = lambda: datetime.now(timezone.utc))
    updated_at = Column(DateTime(timezone = True), onupdate = lambda: datetime.now(timezone.utc))

    datasets: Mapped[List["Dataset"]] = relationship(
        secondary = DatasetData.__tablename__,
        back_populates = "data"
    )

    def to_json(self) -> dict:
        return {
            "id": str(self.id),
            "name": self.name,
            "extension": self.extension,
            "mimeType": self.mime_type,
            "rawDataLocation": self.raw_data_location,
            "createdAt": self.created_at.isoformat(),
            "updatedAt": self.updated_at.isoformat() if self.updated_at else None,
            # "datasets": [dataset.to_json() for dataset in self.datasets]
        }



from duckdb import CatalogException
from cognee.modules.ingestion import discover_directory_datasets
from cognee.modules.data.operations.get_dataset_data import get_dataset_data
from cognee.modules.pipelines.operations.get_pipeline_status import get_pipeline_status
from cognee.infrastructure.databases.relational import get_relational_engine

class datasets():
    @staticmethod
    async def list_datasets():
        db = get_relational_engine()
        return await db.get_datasets()

    @staticmethod
    def discover_datasets(directory_path: str):
        return list(discover_directory_datasets(directory_path).keys())

    @staticmethod
    async def list_data(dataset_id: str, dataset_name: str = None):
        try:
            return await get_dataset_data(dataset_id = dataset_id, dataset_name = dataset_name)
        except CatalogException:
            return None

    @staticmethod
    async def get_status(dataset_ids: list[str]) -> dict:
        try:
            return await get_pipeline_status(dataset_ids)
        except CatalogException:
            return {}

    @staticmethod
    async def delete_dataset(dataset_id: str):
        db = get_relational_engine()
        try:
            return await db.delete_table(dataset_id)
        except CatalogException:
            return {}



import asyncio
from typing import List, Optional
import litellm
from litellm import aembedding
from cognee.infrastructure.databases.vector.embeddings.EmbeddingEngine import EmbeddingEngine

litellm.set_verbose = True

class LiteLLMEmbeddingEngine(EmbeddingEngine):
    api_key: str
    embedding_model: str
    embedding_dimensions: int

    def __init__(
        self,
        embedding_model: Optional[str] = "text-embedding-3-large",
        embedding_dimensions: Optional[int] = 3072,
        api_key: str = None,
    ):
        self.api_key = api_key
        self.embedding_model = embedding_model
        self.embedding_dimensions = embedding_dimensions

    async def embed_text(self, text: List[str]) -> List[List[float]]:
        async def get_embedding(text_):
            response = await aembedding(
                self.embedding_model,
                input = text_,
                api_key = self.api_key
            )

            return response.data[0]["embedding"]

        tasks = [get_embedding(text_) for text_ in text]
        result = await asyncio.gather(*tasks)
        return result

    def get_vector_size(self) -> int:
        return self.embedding_dimensions



from .classify import classify
from .identify import identify
from .save_data_to_file import save_data_to_file
from .get_matched_datasets import get_matched_datasets
from .discover_directory_datasets import discover_directory_datasets



from cognee.modules.users.get_fastapi_users import get_fastapi_users
from cognee.modules.users.models.User import UserRead, UserCreate

def get_register_router():
    return get_fastapi_users().get_register_router(UserRead, UserCreate)






from uuid import uuid4
from .chunk_by_word import chunk_by_word
from cognee.tasks.chunking.chunking_registry import register_chunking_function

@register_chunking_function("sentence")
def chunk_by_sentence(data: str):
    sentence = ""
    paragraph_id = uuid4()
    chunk_index = 0
    word_count = 0

    for (word, word_type) in chunk_by_word(data):
        sentence += (" " if len(sentence) > 0 else "") + word
        word_count += 1

        if word_type == "paragraph_end" or word_type == "sentence_end":
            yield (paragraph_id, chunk_index, sentence, word_count, word_type)
            sentence = ""
            word_count = 0
            paragraph_id = uuid4() if word_type == "paragraph_end" else paragraph_id
            chunk_index = 0 if word_type == "paragraph_end" else chunk_index + 1

    if len(sentence) > 0:
        yield (
            paragraph_id,
            chunk_index,
            sentence,
            word_count,
            "sentence_cut",
        )



from uuid import uuid5, NAMESPACE_OID
from .chunk_by_sentence import chunk_by_sentence
from cognee.tasks.chunking.chunking_registry import register_chunking_function

@register_chunking_function("paragraph")
def chunk_by_paragraph(data: str, paragraph_length: int = 1024, batch_paragraphs = True):
    paragraph = ""
    last_cut_type = None
    last_paragraph_id = None
    paragraph_word_count = 0
    paragraph_chunk_index = 0

    for (paragraph_id, __, sentence, word_count, end_type) in chunk_by_sentence(data):
        if paragraph_word_count > 0 and paragraph_word_count + word_count > paragraph_length:
            if batch_paragraphs is True:
                chunk_id = uuid5(NAMESPACE_OID, paragraph)
                yield dict(
                    text = paragraph.strip(),
                    word_count = paragraph_word_count,
                    id = chunk_id, # When batching paragraphs, the paragraph_id is the same as chunk_id.
                                   # paragraph_id doens't mean anything since multiple paragraphs are merged.
                    chunk_id = chunk_id,
                    chunk_index = paragraph_chunk_index,
                    cut_type = last_cut_type,
                )
            else:
                yield dict(
                    text = paragraph.strip(),
                    word_count = paragraph_word_count,
                    id = last_paragraph_id,
                    chunk_id = uuid5(NAMESPACE_OID, paragraph),
                    chunk_index = paragraph_chunk_index,
                    cut_type = last_cut_type,
                )

            paragraph_chunk_index += 1
            paragraph_word_count = 0
            paragraph = ""

        paragraph += (" " if len(paragraph) > 0 else "") + sentence
        paragraph_word_count += word_count

        if end_type == "paragraph_end" or end_type == "sentence_cut":
            if batch_paragraphs is True:
                paragraph += "\n\n" if end_type == "paragraph_end" else ""
            else:
                yield dict(
                    text = paragraph.strip(),
                    word_count = paragraph_word_count,
                    paragraph_id = paragraph_id,
                    chunk_id = uuid5(NAMESPACE_OID, paragraph),
                    chunk_index = paragraph_chunk_index,
                    cut_type = end_type,
                )

                paragraph_chunk_index = 0
                paragraph_word_count = 0
                paragraph = ""

        last_cut_type = end_type
        last_paragraph_id = paragraph_id

    if len(paragraph) > 0:
        yield dict(
            chunk_id = uuid5(NAMESPACE_OID, paragraph),
            text = paragraph,
            word_count = paragraph_word_count,
            paragraph_id = last_paragraph_id,
            chunk_index = paragraph_chunk_index,
            cut_type = last_cut_type,
        )



from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.llm.prompts import  render_prompt
from cognee.infrastructure.llm.get_llm_client import get_llm_client

async def categorize_relevant_category(query: str, summary, response_model: Type[BaseModel]):
    llm_client = get_llm_client()

    enriched_query= render_prompt("categorize_categories.txt", {"query": query, "categories": summary})


    system_prompt = " Choose the relevant categories and return appropriate output based on the model"

    llm_output = await llm_client.acreate_structured_output(enriched_query, system_prompt, response_model)

    return llm_output.model_dump()



""" This module contains the search function that is used to search for nodes in the graph."""
import asyncio
from enum import Enum
from typing import Dict, Any, Callable, List
from pydantic import BaseModel, field_validator

from cognee.modules.search.graph import search_cypher
from cognee.modules.search.graph.search_adjacent import search_adjacent
from cognee.modules.search.vector.search_traverse import search_traverse
from cognee.modules.search.graph.search_summary import search_summary
from cognee.modules.search.graph.search_similarity import search_similarity
from cognee.shared.utils import send_telemetry
from cognee.modules.users.permissions.methods import get_document_ids_for_user
from cognee.modules.users.methods import get_default_user
from cognee.modules.users.models import User

class SearchType(Enum):
    ADJACENT = "ADJACENT"
    TRAVERSE = "TRAVERSE"
    SIMILARITY = "SIMILARITY"
    SUMMARY = "SUMMARY"
    SUMMARY_CLASSIFICATION = "SUMMARY_CLASSIFICATION"
    NODE_CLASSIFICATION = "NODE_CLASSIFICATION"
    DOCUMENT_CLASSIFICATION = "DOCUMENT_CLASSIFICATION",
    CYPHER = "CYPHER"

    @staticmethod
    def from_str(name: str):
        try:
            return SearchType[name.upper()]
        except KeyError as error:
            raise ValueError(f"{name} is not a valid SearchType") from error

class SearchParameters(BaseModel):
    search_type: SearchType
    params: Dict[str, Any]

    @field_validator("search_type", mode="before")
    def convert_string_to_enum(cls, value): # pylint: disable=no-self-argument
        if isinstance(value, str):
            return SearchType.from_str(value)
        return value


async def search(search_type: str, params: Dict[str, Any], user: User = None) -> List:
    if user is None:
        user = await get_default_user()

    if user is None:
        raise PermissionError("No user found in the system. Please create a user.")

    own_document_ids = await get_document_ids_for_user(user.id)
    search_params = SearchParameters(search_type = search_type, params = params)
    search_results = await specific_search([search_params])

    from uuid import UUID

    filtered_search_results = []

    for search_result in search_results:
        document_id = search_result["document_id"] if "document_id" in search_result else None
        document_id = UUID(document_id) if type(document_id) == str else document_id

        if document_id is None or document_id in own_document_ids:
            filtered_search_results.append(search_result)

    return filtered_search_results


async def specific_search(query_params: List[SearchParameters]) -> List:
    search_functions: Dict[SearchType, Callable] = {
        SearchType.ADJACENT: search_adjacent,
        SearchType.SUMMARY: search_summary,
        SearchType.CYPHER: search_cypher,
        SearchType.TRAVERSE: search_traverse,
        SearchType.SIMILARITY: search_similarity,
    }

    search_tasks = []

    for search_param in query_params:
        search_func = search_functions.get(search_param.search_type)
        if search_func:
            # Schedule the coroutine for execution and store the task
            task = search_func(**search_param.params)
            search_tasks.append(task)

    # Use asyncio.gather to run all scheduled tasks concurrently
    search_results = await asyncio.gather(*search_tasks)

    send_telemetry("cognee.search")

    return search_results[0] if len(search_results) == 1 else search_results



from .ModelBase import Base
from .get_relational_engine import get_relational_engine

async def create_db_and_tables():
    relational_engine = get_relational_engine()

    async with relational_engine.engine.begin() as connection:
        if len(Base.metadata.tables.keys()) > 0:
            await connection.run_sync(Base.metadata.create_all)




import asyncio
from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.databases.vector import get_vector_engine, DataPoint
from cognee.tasks.chunk_extract_summary.models.TextSummary import TextSummary
from cognee.modules.data.extraction.extract_summary import extract_summary
from cognee.modules.data.processing.chunk_types.DocumentChunk import DocumentChunk


async def chunk_extract_summary(data_chunks: list[DocumentChunk], summarization_model: Type[BaseModel], collection_name: str = "summaries"):
    if len(data_chunks) == 0:
        return data_chunks

    chunk_summaries = await asyncio.gather(
        *[extract_summary(chunk.text, summarization_model) for chunk in data_chunks]
    )

    vector_engine = get_vector_engine()

    await vector_engine.create_collection(collection_name, payload_schema=TextSummary)

    await vector_engine.create_data_points(
        collection_name,
        [
            DataPoint[TextSummary](
                id = str(chunk.chunk_id),
                payload = dict(
                    chunk_id = str(chunk.chunk_id),
                    document_id = str(chunk.document_id),
                    text = chunk_summaries[chunk_index].summary,
                ),
                embed_field = "text",
            ) for (chunk_index, chunk) in enumerate(data_chunks)
        ],
    )

    return data_chunks



import string
import random
from typing import BinaryIO, Union
from cognee.base_config import get_base_config
from cognee.infrastructure.files.storage import LocalStorage
from .classify import classify

def save_data_to_file(data: Union[str, BinaryIO], dataset_name: str, filename: str = None):
    base_config = get_base_config()
    data_directory_path = base_config.data_root_directory

    classified_data = classify(data, filename)

    storage_path = data_directory_path + "/" + dataset_name.replace(".", "/")
    LocalStorage.ensure_directory_exists(storage_path)

    file_metadata = classified_data.get_metadata()
    if "name" not in file_metadata or file_metadata["name"] is None:
        letters = string.ascii_lowercase
        random_string = "".join(random.choice(letters) for _ in range(32))
        file_metadata["name"] = "text_" + random_string + ".txt"
    file_name = file_metadata["name"]
    LocalStorage(storage_path).store(file_name, classified_data.get_data())

    return "file://" + storage_path + "/" + file_name



import duckdb
import os
class DuckDBAdapter():
    def __init__(self, db_path: str, db_name: str):
        self.db_location = os.path.abspath(os.path.join(db_path, db_name))

        self.get_connection = lambda: duckdb.connect(self.db_location)

    def get_datasets(self):
        with self.get_connection() as connection:
            tables = connection.sql("SELECT DISTINCT schema_name FROM duckdb_tables();").to_df().to_dict("list")

        return list(
            filter(
                lambda schema_name: not schema_name.endswith("staging") and schema_name != "cognee",
                tables["schema_name"]
            )
        )

    def get_files_metadata(self, dataset_name: str):
        with self.get_connection() as connection:
            return connection.sql(f"SELECT id, name, file_path, extension, mime_type FROM {dataset_name}.file_metadata;").to_df().to_dict("records")

    def create_table(self, schema_name: str, table_name: str, table_config: list[dict]):
        fields_query_parts = []

        for table_config_item in table_config:
            fields_query_parts.append(f"{table_config_item['name']} {table_config_item['type']}")

        with self.get_connection() as connection:
            query = f"CREATE SCHEMA IF NOT EXISTS {schema_name};"
            connection.execute(query)

        with self.get_connection() as connection:
            query = f"CREATE TABLE IF NOT EXISTS {schema_name}.{table_name} ({', '.join(fields_query_parts)});"
            connection.execute(query)

    def delete_table(self, table_name: str):
        with self.get_connection() as connection:
            query = f"DROP TABLE IF EXISTS {table_name};"
            connection.execute(query)

    def insert_data(self, schema_name: str, table_name: str, data: list[dict]):
        def get_values(data_entry: list):
            return ", ".join([f"'{value}'" if isinstance(value, str) else value for value in data_entry])
      
        columns = ", ".join(data[0].keys())
        values = ", ".join([f"({get_values(data_entry.values())})" for data_entry in data])

        with self.get_connection() as connection:
            query = f"INSERT INTO {schema_name}.{table_name} ({columns}) VALUES {values};"
            connection.execute(query)

    def get_data(self, table_name: str, filters: dict = None):
        with self.get_connection() as connection:
            def get_values(values: list):
                return ", ".join([f"'{value}'" for value in values])

            def get_filters(filters: dict):
                return " AND ".join([
                    f"{key} IN ({get_values(value)})" if isinstance(value, list)
                    else f"{key} = '{value}'" for (key, value) in filters.items()
                ])

            query = f"SELECT * FROM {table_name}" + (";" if filters is None else f" WHERE {get_filters(filters)};")
            results = connection.sql(query).to_df().to_dict("records")

            return {
                result["data_id"]: result["status"] for result in results
            }

    def execute_query(self, query):
        with self.get_connection() as connection:
            return connection.sql(query).to_df().to_dict("records")

    def load_cognify_data(self, data):
        with self.get_connection() as connection:
            # Ensure the "cognify" table exists
            connection.execute("""
                CREATE TABLE IF NOT EXISTS cognify (
                    document_id STRING,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    updated_at TIMESTAMP DEFAULT NULL,
                    processed BOOLEAN DEFAULT FALSE,
                    document_id_target STRING NULL
                );
            """)

        # Prepare the insert statement
        insert_query = """
            INSERT INTO cognify (document_id)
            VALUES (?);
        """

        # Insert each record into the "cognify" table
        for record in data:
            with self.get_connection() as connection:
                connection.execute(insert_query, [
                    record.get("document_id"),
                ])

    def fetch_cognify_data(self, excluded_document_id: str):
        # SQL command to create the "cognify" table with the specified columns
        create_table_sql = """
        CREATE TABLE IF NOT EXISTS cognify (
            document_id STRING,
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP DEFAULT NULL,
            processed BOOLEAN DEFAULT FALSE,
            document_id_target STRING NULL
        );
        """
        with self.get_connection() as connection:
            # Execute the SQL command to create the table
            connection.execute(create_table_sql)

        # SQL command to select data from the "cognify" table
        select_data_sql = f"SELECT document_id, created_at, updated_at, processed FROM cognify WHERE document_id != '{excluded_document_id}' AND processed = FALSE;"

        with self.get_connection() as connection:
            # Execute the query and fetch the results
            records = connection.sql(select_data_sql).to_df().to_dict("records")

        # If records are fetched, update the "processed" column to "True"
        if records:
            # Fetching document_ids from the records to update the "processed" column
            document_ids = tuple(record["document_id"] for record in records)
            # SQL command to update the "processed" column to "True" for fetched records
            update_data_sql = f"UPDATE cognify SET processed = TRUE WHERE document_id IN {document_ids};"

            with self.get_connection() as connection:
                # Execute the update query
                connection.execute(update_data_sql)

        # Return the fetched records
        return records


    def delete_cognify_data(self):
        # SQL command to create the "cognify" table with the specified columns
        create_table_sql = """
        CREATE TABLE IF NOT EXISTS cognify (
            document_id STRING,
            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
            updated_at TIMESTAMP DEFAULT NULL,
            processed BOOLEAN DEFAULT FALSE,
            document_id_target STRING NULL
        );
        """

        with self.get_connection() as connection:
            # Execute the SQL command to create the table
            connection.execute(create_table_sql)

        with self.get_connection() as connection:
            # SQL command to select data from the "cognify" table
            select_data_sql = "DELETE FROM cognify;"
            connection.sql(select_data_sql)
            drop_data_sql = "DROP TABLE cognify;"
            connection.sql(drop_data_sql)

    def delete_database(self):
        from cognee.infrastructure.files.storage import LocalStorage

        if LocalStorage.file_exists(self.db_location):
            LocalStorage.remove(self.db_location)

        if LocalStorage.file_exists(self.db_location + ".wal"):
            LocalStorage.remove(self.db_location + ".wal")



import asyncio
import base64
import os
from pathlib import Path
from typing import List, Type

import aiofiles
import openai
import instructor
from pydantic import BaseModel
from tenacity import retry, stop_after_attempt

from cognee.base_config import get_base_config
from cognee.infrastructure.llm.llm_interface import LLMInterface
from cognee.infrastructure.llm.prompts import read_query_prompt
from cognee.shared.data_models import MonitoringTool
import logging
logging.basicConfig(level=logging.DEBUG)

class OpenAIAdapter(LLMInterface):
    name = "OpenAI"
    model: str
    api_key: str
  
    """Adapter for OpenAI's GPT-3, GPT=4 API"""
    def __init__(self, api_key: str, model: str, transcription_model:str, streaming: bool = False):
        base_config = get_base_config()

        # if base_config.monitoring_tool == MonitoringTool.LANGFUSE:
        #     from langfuse.openai import AsyncOpenAI, OpenAI
        # elif base_config.monitoring_tool == MonitoringTool.LANGSMITH:
        #     from langsmith import wrappers
        #     from openai import AsyncOpenAI
        #     AsyncOpenAI = wrappers.wrap_openai(AsyncOpenAI())
        # else:
        from openai import AsyncOpenAI, OpenAI

        self.aclient = instructor.from_openai(AsyncOpenAI(api_key = api_key))
        self.client = instructor.from_openai(OpenAI(api_key = api_key))
        self.base_openai_client = OpenAI(api_key = api_key)
        self.transcription_model = "whisper-1"
        self.model = model
        self.api_key = api_key
        self.streaming = streaming
    @retry(stop = stop_after_attempt(5))
    def completions_with_backoff(self, **kwargs):
        """Wrapper around ChatCompletion.create w/ backoff"""
        return openai.chat.completions.create(**kwargs)

    @retry(stop = stop_after_attempt(5))
    async def acompletions_with_backoff(self,**kwargs):
        """Wrapper around ChatCompletion.acreate w/ backoff"""
        return await openai.chat.completions.acreate(**kwargs)

    @retry(stop = stop_after_attempt(5))
    async def acreate_embedding_with_backoff(self, input: List[str], model: str = "text-embedding-3-large"):
        """Wrapper around Embedding.acreate w/ backoff"""

        return await self.aclient.embeddings.create(input = input, model = model)

    async def async_get_embedding_with_backoff(self, text, model="text-embedding-3-large"):
        """To get text embeddings, import/call this function
        It specifies defaults + handles rate-limiting + is async"""
        text = text.replace("\n", " ")
        response = await self.aclient.embeddings.create(input = text, model = model)
        embedding = response.data[0].embedding
        return embedding

    @retry(stop = stop_after_attempt(5))
    def create_embedding_with_backoff(self, **kwargs):
        """Wrapper around Embedding.create w/ backoff"""
        return openai.embeddings.create(**kwargs)

    def get_embedding_with_backoff(self, text: str, model: str = "text-embedding-3-large"):
        """To get text embeddings, import/call this function
        It specifies defaults + handles rate-limiting
        :param text: str
        :param model: str
        """
        text = text.replace("\n", " ")
        response = self.create_embedding_with_backoff(input=[text], model=model)
        embedding = response.data[0].embedding
        return embedding

    async def async_get_batch_embeddings_with_backoff(self, texts: List[str], models: List[str]):
        """To get multiple text embeddings in parallel, import/call this function
        It specifies defaults + handles rate-limiting + is async"""
        # Collect all coroutines
        coroutines = (self.async_get_embedding_with_backoff(text, model)
            for text, model in zip(texts, models))

        # Run the coroutines in parallel and gather the results
        embeddings = await asyncio.gather(*coroutines)

        return embeddings

    @retry(stop = stop_after_attempt(5))
    async def acreate_structured_output(self, text_input: str, system_prompt: str, response_model: Type[BaseModel]) -> BaseModel:
        """Generate a response from a user query."""

        return await self.aclient.chat.completions.create(
            model = self.model,
            messages = [
                {
                    "role": "user",
                    "content": f"""Use the given format to
                    extract information from the following input: {text_input}. """,
                },
                {"role": "system", "content": system_prompt},
            ],
            response_model = response_model,
        )


    @retry(stop = stop_after_attempt(5))
    def create_structured_output(self, text_input: str, system_prompt: str, response_model: Type[BaseModel]) -> BaseModel:
        """Generate a response from a user query."""

        return self.client.chat.completions.create(
            model = self.model,
            messages = [
                {
                    "role": "user",
                    "content": f"""Use the given format to
                    extract information from the following input: {text_input}. """,
                },
                {"role": "system", "content": system_prompt},
            ],
            response_model = response_model,
        )

    @retry(stop = stop_after_attempt(5))
    def create_transcript(self, input):
        """Generate a audio transcript from a user query."""

        if not os.path.isfile(input):
            raise FileNotFoundError(f"The file {input} does not exist.")

        with open(input, 'rb') as audio_file:
            audio_data = audio_file.read()



        transcription = self.base_openai_client.audio.transcriptions.create(
                  model=self.transcription_model ,
                  file=Path(input),
                )

        return transcription


    @retry(stop = stop_after_attempt(5))
    def transcribe_image(self, input) -> BaseModel:
        with open(input, "rb") as image_file:
            encoded_image = base64.b64encode(image_file.read()).decode('utf-8')

        return self.base_openai_client.chat.completions.create(
            model=self.model,
            messages=[
                        {
                          "role": "user",
                          "content": [
                            {"type": "text", "text": "What’s in this image?"},
                            {
                              "type": "image_url",
                              "image_url": {
                                "url": f"data:image/jpeg;base64,{encoded_image}",
                              },
                            },
                          ],
                        }
                      ],
            max_tokens=300,
        )
    def show_prompt(self, text_input: str, system_prompt: str) -> str:
        """Format and display the prompt for a user query."""
        if not text_input:
            text_input = "No user input provided."
        if not system_prompt:
            raise ValueError("No system prompt path provided.")
        system_prompt = read_query_prompt(system_prompt)

        formatted_prompt = f"""System Prompt:\n{system_prompt}\n\nUser Input:\n{text_input}\n""" if system_prompt else None
        return formatted_prompt



from typing import BinaryIO
from cognee.infrastructure.data.utils.extract_keywords import extract_keywords
from .IngestionData import IngestionData

def create_text_data(data: str):
    return TextData(data)

class TextData(IngestionData):
    data: str = None
    metadata = None

    def __init__(self, data: BinaryIO):
        self.data = data

    def get_identifier(self):
        keywords = extract_keywords(self.data)

        return "text/plain" + "_" + "|".join(keywords)

    def get_metadata(self):
        self.ensure_metadata()

        return self.metadata

    def ensure_metadata(self):
        if self.metadata is None:
            self.metadata = {}

    def get_data(self):
        return self.data



import asyncio
import nest_asyncio
import dspy
from cognee.modules.search.vector.search_similarity import search_similarity

nest_asyncio.apply()

class AnswerFromContext(dspy.Signature):
    question: str = dspy.InputField()
    context: str = dspy.InputField(desc = "Context to use for answer generation.")
    answer: str = dspy.OutputField()

question_answer_llm = dspy.OpenAI(model = "gpt-3.5-turbo-instruct")

class CogneeSearch(dspy.Module):
    def __init__(self, ):
        super().__init__()
        self.generate_answer = dspy.TypedChainOfThought(AnswerFromContext)

    def forward(self, question):
        context = asyncio.run(search_similarity(question))

        context_text = "\n".join(context)
        print(f"Context: {context_text}")

        with dspy.context(lm = question_answer_llm):
            answer_prediction = self.generate_answer(context = context_text, question = question)
            answer = answer_prediction.answer

            print(f"Question: {question}")
            print(f"Answer: {answer}")

        return dspy.Prediction(context = context_text, answer = answer)



import re
from nltk.downloader import download
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import sent_tokenize, word_tokenize
from nltk.corpus import stopwords, wordnet
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.decomposition import TruncatedSVD

def extract_topics(text: str):
    sentences = sent_tokenize(text)

    try:
        wordnet.ensure_loaded()
    except LookupError:
        download("wordnet")
        wordnet.ensure_loaded()

    lemmatizer = WordNetLemmatizer()
    base_notation_sentences = [lemmatizer.lemmatize(sentence) for sentence in sentences]

    tf_vectorizer = TfidfVectorizer(tokenizer = word_tokenize, token_pattern = None)
    transformed_corpus = tf_vectorizer.fit_transform(base_notation_sentences)

    svd = TruncatedSVD(n_components = 10)
    svd_corpus = svd.fit(transformed_corpus)

    feature_scores = dict(
        zip(
            tf_vectorizer.vocabulary_,
            svd_corpus.components_[0]
        )
    )

    topics = sorted(
        feature_scores,
        # key = feature_scores.get,
        key = lambda x: transformed_corpus[0, tf_vectorizer.vocabulary_[x]],
        reverse = True,
    )[:10]

    return topics


def clean_text(text: str):
    text = re.sub(r"[ \t]{2,}|[\n\r]", " ", text.lower())
    return re.sub(r"[`\"'.,;!?…]", "", text).strip()

def remove_stop_words(text: str):
    try:
        stopwords.ensure_loaded()
    except LookupError:
        download("stopwords")
        stopwords.ensure_loaded()

    stop_words = set(stopwords.words("english"))
    text = text.split()
    text = [word for word in text if not word in stop_words]
    return " ".join(text)


if __name__ == "__main__":
    text = """Lorem Ipsum is simply dummy text of the printing and typesetting industry... Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type and scrambled it to make a type specimen book… It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged. It was popularised in the 1960s with the release of Letraset sheets containing Lorem Ipsum passages, and more recently with desktop publishing software like Aldus PageMaker including versions of Lorem Ipsum.
        Why do we use it?
        It is a long established fact that a reader will be distracted by the readable content of a page when looking at its layout! The point of using Lorem Ipsum is that it has a more-or-less normal distribution of letters, as opposed to using 'Content here, content here', making it look like readable English. Many desktop publishing packages and web page editors now use Lorem Ipsum as their default model text, and a search for 'lorem ipsum' will uncover many web sites still in their infancy. Various versions have evolved over the years, sometimes by accident, sometimes on purpose (injected humour and the like).
    """
    print(extract_topics(remove_stop_words(clean_text(text))))



import asyncio
import logging
from typing import List, Optional
from ..vector_db_interface import VectorDBInterface
from ..models.DataPoint import DataPoint
from ..models.ScoredResult import ScoredResult
from ..embeddings.EmbeddingEngine import EmbeddingEngine

logger = logging.getLogger("WeaviateAdapter")

class WeaviateAdapter(VectorDBInterface):
    name = "Weaviate"
    url: str
    api_key: str
    embedding_engine: EmbeddingEngine = None

    def __init__(self, url: str, api_key: str, embedding_engine: EmbeddingEngine):
        import weaviate
        import weaviate.classes as wvc

        self.url = url
        self.api_key = api_key

        self.embedding_engine = embedding_engine

        self.client = weaviate.connect_to_wcs(
            cluster_url = url,
            auth_credentials = weaviate.auth.AuthApiKey(api_key),
            additional_config = wvc.init.AdditionalConfig(timeout = wvc.init.Timeout(init=30))
        )

    async def embed_data(self, data: List[str]) -> List[float]:
        return await self.embedding_engine.embed_text(data)

    async def has_collection(self, collection_name: str) -> bool:
        future = asyncio.Future()

        future.set_result(self.client.collections.exists(collection_name))

        return await future

    async def create_collection(
        self,
        collection_name: str,
        payload_schema = None,
    ):
        import weaviate.classes.config as wvcc

        future = asyncio.Future()

        future.set_result(
            self.client.collections.create(
                name=collection_name,
                properties=[
                    wvcc.Property(
                        name="text",
                        data_type=wvcc.DataType.TEXT,
                        skip_vectorization=True
                    )
                ]
            )
        )

        return await future

    def get_collection(self, collection_name: str):
        return self.client.collections.get(collection_name)

    async def create_data_points(self, collection_name: str, data_points: List[DataPoint]):
        from weaviate.classes.data import DataObject

        data_vectors = await self.embed_data(
            list(map(lambda data_point: data_point.get_embeddable_data(), data_points)))

        def convert_to_weaviate_data_points(data_point: DataPoint):
            vector = data_vectors[data_points.index(data_point)]
            return DataObject(
                uuid = data_point.id,
                properties = data_point.payload.dict(),
                vector = vector
            )

        data_points = list(map(convert_to_weaviate_data_points, data_points))

        collection = self.get_collection(collection_name)

        try:
            if len(data_points) > 1:
                return collection.data.insert_many(data_points)
            else:
                return collection.data.insert(data_points[0])
            # with collection.batch.dynamic() as batch:
            #     for point in data_points:
            #         batch.add_object(
            #             uuid = point.uuid,
            #             properties = point.properties,
            #             vector = point.vector
            #         )
        except Exception as error:
            logger.error("Error creating data points: %s", str(error))
            raise error

    async def retrieve(self, collection_name: str, data_point_ids: list[str]):
        from weaviate.classes.query import Filter
        future = asyncio.Future()

        data_points = self.get_collection(collection_name).query.fetch_objects(
            filters = Filter.by_id().contains_any(data_point_ids)
        )

        for data_point in data_points:
            data_point.payload = data_point.properties
            del data_point.properties

        future.set_result(data_points)

        return await future

    async def search(
            self,
            collection_name: str,
            query_text: Optional[str] = None,
            query_vector: Optional[List[float]] = None,
            limit: int = None,
            with_vector: bool = False
    ):
        import weaviate.classes as wvc

        if query_text is None and query_vector is None:
            raise ValueError("One of query_text or query_vector must be provided!")

        if query_vector is None:
            query_vector = (await self.embed_data([query_text]))[0]

        search_result = self.get_collection(collection_name).query.hybrid(
            query = None,
            vector = query_vector,
            limit = limit,
            include_vector = with_vector,
            return_metadata = wvc.query.MetadataQuery(score=True),
        )

        return [
            ScoredResult(
                id=str(result.uuid),
                payload=result.properties,
                score=float(result.metadata.score)
            ) for result in search_result.objects
        ]

    async def batch_search(self, collection_name: str, query_texts: List[str], limit: int, with_vectors: bool = False):
        def query_search(query_vector):
            return self.search(collection_name, query_vector=query_vector, limit=limit, with_vector=with_vectors)

        return [await query_search(query_vector) for query_vector in await self.embed_data(query_texts)]
        
    async def delete_data_points(self, collection_name: str, data_point_ids: list[str]):
        from weaviate.classes.query import Filter
        future = asyncio.Future()

        result = self.get_collection(collection_name).data.delete_many(
            filters = Filter.by_id().contains_any(data_point_ids)
        )
        future.set_result(result)

        return await future

    async def prune(self):
        self.client.collections.delete_all()



from cognee.modules.users.methods import create_user as create_user_method


async def create_user(email: str, password: str, is_superuser: bool = False):
    user = await create_user_method(
        email = email,
        password = password,
        is_superuser = is_superuser,
        is_verified = True,
    )

    return user



def generate_node_name(name: str) -> str:
    return name.lower().replace(" ", "_").replace("'", "")

def generate_node_id(node_id: str) -> str:
    return node_id.lower().replace(" ", "_").replace("'", "")



import asyncio
from cognee.shared.utils import render_graph
from cognee.infrastructure.databases.graph import get_graph_engine

if __name__ == "__main__":
    async def main():
        import os
        import pathlib
        import cognee

        data_directory_path = str(pathlib.Path(os.path.join(pathlib.Path(__file__).parent, ".data_storage/test_library")).resolve())
        cognee.config.data_root_directory(data_directory_path)
        cognee_directory_path = str(pathlib.Path(os.path.join(pathlib.Path(__file__).parent, ".cognee_system/test_library")).resolve())
        cognee.config.system_root_directory(cognee_directory_path)

        graph_client = await get_graph_engine()
        graph = graph_client.graph

        graph_url = await render_graph(graph)

        print(graph_url)

    asyncio.run(main())



from .discover_directory_datasets import discover_directory_datasets

def get_matched_datasets(data_path: str, dataset_name_to_match: str = None):
    datasets = discover_directory_datasets(data_path)

    matched_datasets = []

    for dataset_name, dataset_files in datasets.items():
        if dataset_name_to_match is None or dataset_name.startswith(dataset_name_to_match):
            matched_datasets.append([dataset_name, dataset_files])

    return matched_datasets



from .chunk_extract_summary.chunk_extract_summary import chunk_extract_summary
from .chunk_naive_llm_classifier.chunk_naive_llm_classifier import chunk_naive_llm_classifier
from .chunk_remove_disconnected.chunk_remove_disconnected import chunk_remove_disconnected
from .chunk_update_check.chunk_update_check import chunk_update_check
from .chunks_into_graph.chunks_into_graph import chunks_into_graph
from .save_chunks_to_store.save_chunks_to_store import save_chunks_to_store
from .source_documents_to_chunks.source_documents_to_chunks import source_documents_to_chunks
from .infer_data_ontology.infer_data_ontology import infer_data_ontology
from .check_permissions_on_documents.check_permissions_on_documents import check_permissions_on_documents
from .classify_documents.classify_documents import classify_documents



import os
from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict
from cognee.root_dir import get_absolute_path

class RelationalConfig(BaseSettings):
    db_path: str =  os.path.join(get_absolute_path(".cognee_system"), "databases")
    db_name: str =  "cognee_db"
    db_host: str =  "localhost"
    db_port: str =  "5432"
    db_user: str = "cognee"
    db_password: str =  "cognee"
    db_provider: str = "postgresql+asyncpg"
    db_file_path: str = os.path.join(db_path, db_name)


    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "db_path": self.db_path,
            "db_name": self.db_name,
            "db_host": self.db_host,
            "db_port": self.db_port,
            "db_user": self.db_user,
            "db_password": self.db_password,
            "db_provider": self.db_provider,
        }

@lru_cache
def get_relational_config():
    return RelationalConfig()



class HaystackChunkEngine:
    def __init__(self, chunk_strategy=None, source_data=None, chunk_size=None, chunk_overlap=None):
        self.chunk_strategy = chunk_strategy
        self.source_data = source_data
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap



from typing import Optional
from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict

class LLMConfig(BaseSettings):
    llm_provider: str = "openai"
    llm_model: str = "gpt-4o"
    llm_endpoint: str = ""
    llm_api_key: Optional[str] = None
    llm_temperature: float = 0.0
    llm_streaming: bool = False
    transcription_model: str = "whisper-1"

    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "provider": self.llm_provider,
            "model": self.llm_model,
            "endpoint": self.llm_endpoint,
            "apiKey": self.llm_api_key,
            "temperature": self.llm_temperature,
            "streaming": self.llm_stream,
            "transcriptionModel": self.transcription_model
        }

@lru_cache
def get_llm_config():
    return LLMConfig()



import logging
from typing import List, Dict, Optional
from qdrant_client import AsyncQdrantClient, models
from ..vector_db_interface import VectorDBInterface
from ..models.DataPoint import DataPoint
from ..embeddings.EmbeddingEngine import EmbeddingEngine

logger = logging.getLogger("QDrantAdapter")

# class CollectionConfig(BaseModel, extra = "forbid"):
#     vector_config: Dict[str, models.VectorParams] = Field(..., description="Vectors configuration" )
#     hnsw_config: Optional[models.HnswConfig] = Field(default = None, description="HNSW vector index configuration")
#     optimizers_config: Optional[models.OptimizersConfig] = Field(default = None, description="Optimizers configuration")
#     quantization_config: Optional[models.QuantizationConfig] = Field(default = None, description="Quantization configuration")

def create_hnsw_config(hnsw_config: Dict):
    if hnsw_config is not None:
        return models.HnswConfig()
    return None

def create_optimizers_config(optimizers_config: Dict):
    if optimizers_config is not None:
        return models.OptimizersConfig()
    return None

def create_quantization_config(quantization_config: Dict):
    if quantization_config is not None:
        return models.QuantizationConfig()
    return None

class QDrantAdapter(VectorDBInterface):
    name = "Qdrant"
    url: str = None
    api_key: str = None
    qdrant_path: str = None

    def __init__(self, url, api_key, embedding_engine: EmbeddingEngine, qdrant_path = None):
        self.embedding_engine = embedding_engine

        if qdrant_path is not None:
            self.qdrant_path = qdrant_path
        else:
            self.url = url
            self.api_key = api_key

    def get_qdrant_client(self) -> AsyncQdrantClient:
        if self.qdrant_path is not None:
            return AsyncQdrantClient(
                path = self.qdrant_path, port=6333
            )
        elif self.url is not None:
            return AsyncQdrantClient(
                url = self.url,
                api_key = self.api_key,
                port = 6333
            )

        return AsyncQdrantClient(
            location = ":memory:"
        )

    async def embed_data(self, data: List[str]) -> List[float]:
        return await self.embedding_engine.embed_text(data)

    async def has_collection(self, collection_name: str) -> bool:
        client = self.get_qdrant_client()
        result = await client.collection_exists(collection_name)
        await client.close()
        return result

    async def create_collection(
      self,
      collection_name: str,
      payload_schema = None,
    ):
        client = self.get_qdrant_client()

        result = await client.create_collection(
            collection_name = collection_name,
            vectors_config = {
                "text": models.VectorParams(
                    size = self.embedding_engine.get_vector_size(),
                    distance = "Cosine"
                )
            }
        )

        await client.close()

        return result

    async def create_data_points(self, collection_name: str, data_points: List[DataPoint]):
        client = self.get_qdrant_client()

        data_vectors = await self.embed_data([data_point.get_embeddable_data() for data_point in data_points])

        def convert_to_qdrant_point(data_point: DataPoint):
            return models.PointStruct(
                id = data_point.id,
                payload = data_point.payload.dict(),
                vector = {
                    "text": data_vectors[data_points.index(data_point)]
                }
            )

        points = [convert_to_qdrant_point(point) for point in data_points]

        try:
            result = await client.upload_points(
                collection_name = collection_name,
                points = points
            )
            return result
        except Exception as error:
            logger.error("Error uploading data points to Qdrant: %s", str(error))
            raise error
        finally:
            await client.close()

    async def retrieve(self, collection_name: str, data_point_ids: list[str]):
        client = self.get_qdrant_client()
        results = await client.retrieve(collection_name, data_point_ids, with_payload = True)
        await client.close()
        return results

    async def search(
        self,
        collection_name: str,
        query_text: Optional[str] = None,
        query_vector: Optional[List[float]] = None,
        limit: int = 5,
        with_vector: bool = False
    ):
        if query_text is None and query_vector is None:
            raise ValueError("One of query_text or query_vector must be provided!")

        client = self.get_qdrant_client()

        result = await client.search(
            collection_name = collection_name,
            query_vector = models.NamedVector(
                name = "text",
                vector = query_vector if query_vector is not None else (await self.embed_data([query_text]))[0],
            ),
            limit = limit,
            with_vectors = with_vector
        )

        await client.close()

        return result


    async def batch_search(self, collection_name: str, query_texts: List[str], limit: int = None, with_vectors: bool = False):
        """
        Perform batch search in a Qdrant collection with dynamic search requests.

        Args:
        - collection_name (str): Name of the collection to search in.
        - query_texts (List[str]): List of query texts to search for.
        - limit (int): List of result limits for search requests.
        - with_vectors (bool, optional): Bool indicating whether to return vectors for search requests.

        Returns:
        - results: The search results from Qdrant.
        """

        vectors = await self.embed_data(query_texts)

        # Generate dynamic search requests based on the provided embeddings
        requests = [
            models.SearchRequest(
                vector = models.NamedVector(
                    name = "text",
                    vector = vector
                ),
                limit = limit,
                with_vector = with_vectors
            ) for vector in vectors
        ]

        client = self.get_qdrant_client()

        # Perform batch search with the dynamically generated requests
        results = await client.search_batch(
            collection_name = collection_name,
            requests = requests
        )

        await client.close()

        return [filter(lambda result: result.score > 0.9, result_group) for result_group in results]

    async def delete_data_points(self, collection_name: str, data_point_ids: list[str]):
        client = self.get_qdrant_client()
        results = await client.delete(collection_name, data_point_ids)
        return results

    async def prune(self):
        client = self.get_qdrant_client()

        response = await client.get_collections()

        for collection in response.collections:
            await client.delete_collection(collection.name)

        await client.close()



import json
import asyncio
from uuid import uuid5, NAMESPACE_OID
from datetime import datetime, timezone
from typing import Type
from pydantic import BaseModel
from cognee.infrastructure.databases.graph import get_graph_engine
from cognee.infrastructure.databases.vector import DataPoint, get_vector_engine
from cognee.modules.data.extraction.knowledge_graph.extract_content_graph import extract_content_graph
from cognee.modules.data.processing.chunk_types.DocumentChunk import DocumentChunk
from cognee.modules.graph.utils import generate_node_id, generate_node_name


class EntityNode(BaseModel):
    uuid: str
    name: str
    type: str
    description: str
    created_at: datetime
    updated_at: datetime

async def chunks_into_graph(data_chunks: list[DocumentChunk], graph_model: Type[BaseModel], collection_name: str):
    chunk_graphs = await asyncio.gather(
        *[extract_content_graph(chunk.text, graph_model) for chunk in data_chunks]
    )

    vector_engine = get_vector_engine()
    graph_engine = await get_graph_engine()

    has_collection = await vector_engine.has_collection(collection_name)

    if not has_collection:
        await vector_engine.create_collection(collection_name, payload_schema = EntityNode)

    processed_nodes = {}
    type_node_edges = []
    entity_node_edges = []
    type_entity_edges = []

    for (chunk_index, chunk) in enumerate(data_chunks):
        chunk_graph = chunk_graphs[chunk_index]
        for node in chunk_graph.nodes:
            type_node_id = generate_node_id(node.type)
            entity_node_id = generate_node_id(node.id)

            if type_node_id not in processed_nodes:
                type_node_edges.append((str(chunk.chunk_id), type_node_id, "contains_entity_type"))
                processed_nodes[type_node_id] = True

            if entity_node_id not in processed_nodes:
                entity_node_edges.append((str(chunk.chunk_id), entity_node_id, "contains_entity"))
                type_entity_edges.append((entity_node_id, type_node_id, "is_entity_type"))
                processed_nodes[entity_node_id] = True

        graph_node_edges = [
            (edge.source_node_id, edge.target_node_id, edge.relationship_name) \
                        for edge in chunk_graph.edges
        ]

    existing_edges = await graph_engine.has_edges([
        *type_node_edges,
        *entity_node_edges,
        *type_entity_edges,
        *graph_node_edges,
    ])

    existing_edges_map = {}
    existing_nodes_map = {}

    for edge in existing_edges:
        existing_edges_map[edge[0] + edge[1] + edge[2]] = True
        existing_nodes_map[edge[0]] = True

    graph_nodes = []
    graph_edges = []
    data_points = []

    for (chunk_index, chunk) in enumerate(data_chunks):
        graph = chunk_graphs[chunk_index]
        if graph is None:
            continue

        for node in graph.nodes:
            node_id = generate_node_id(node.id)
            node_name = generate_node_name(node.name)

            type_node_id = generate_node_id(node.type)
            type_node_name = generate_node_name(node.type)

            if node_id not in existing_nodes_map:
                node_data = dict(
                    uuid = node_id,
                    name = node_name,
                    type = node_name,
                    description = node.description,
                    created_at = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                    updated_at = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                )

                graph_nodes.append((
                    node_id,
                    dict(
                        **node_data,
                        properties = json.dumps(node.properties),
                    )
                ))

                data_points.append(DataPoint[EntityNode](
                    id = str(uuid5(NAMESPACE_OID, node_id)),
                    payload = node_data,
                    embed_field = "name",
                ))

                existing_nodes_map[node_id] = True

            edge_key = str(chunk.chunk_id) + node_id + "contains_entity"

            if edge_key not in existing_edges_map:
                graph_edges.append((
                    str(chunk.chunk_id),
                    node_id,
                    "contains_entity",
                    dict(
                        relationship_name = "contains_entity",
                        source_node_id = str(chunk.chunk_id),
                        target_node_id = node_id,
                    ),
                ))

                # Add relationship between entity type and entity itself: "Jake is Person"
                graph_edges.append((
                    node_id,
                    type_node_id,
                    "is_entity_type",
                    dict(
                        relationship_name = "is_entity_type",
                        source_node_id = type_node_id,
                        target_node_id = node_id,
                    ),
                ))

                existing_edges_map[edge_key] = True

            if type_node_id not in existing_nodes_map:
                type_node_data = dict(
                    uuid = type_node_id,
                    name = type_node_name,
                    type = type_node_id,
                    description = type_node_name,
                    created_at = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                    updated_at = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                )

                graph_nodes.append((type_node_id, dict(
                    **type_node_data,
                    properties = json.dumps(node.properties)
                )))

                data_points.append(DataPoint[EntityNode](
                    id = str(uuid5(NAMESPACE_OID, type_node_id)),
                    payload = type_node_data,
                    embed_field = "name",
                ))

                existing_nodes_map[type_node_id] = True

            edge_key = str(chunk.chunk_id) + type_node_id + "contains_entity_type"

            if edge_key not in existing_edges_map:
                graph_edges.append((
                    str(chunk.chunk_id),
                    type_node_id,
                    "contains_entity_type",
                    dict(
                        relationship_name = "contains_entity_type",
                        source_node_id = str(chunk.chunk_id),
                        target_node_id = type_node_id,
                    ),
                ))

                existing_edges_map[edge_key] = True

            # Add relationship that came from graphs.
            for edge in graph.edges:
                source_node_id = generate_node_id(edge.source_node_id)
                target_node_id = generate_node_id(edge.target_node_id)
                relationship_name = generate_node_name(edge.relationship_name)
                edge_key = source_node_id + target_node_id + relationship_name

                if edge_key not in existing_edges_map:
                    graph_edges.append((
                        generate_node_id(edge.source_node_id),
                        generate_node_id(edge.target_node_id),
                        edge.relationship_name,
                        dict(
                            relationship_name = generate_node_name(edge.relationship_name),
                            source_node_id = generate_node_id(edge.source_node_id),
                            target_node_id = generate_node_id(edge.target_node_id),
                            properties = json.dumps(edge.properties),
                        ),
                    ))
                    existing_edges_map[edge_key] = True

    if len(data_points) > 0:
        await vector_engine.create_data_points(collection_name, data_points)

    if len(graph_nodes) > 0:
        await graph_engine.add_nodes(graph_nodes)

    if len(graph_edges) > 0:
        await graph_engine.add_edges(graph_edges)

    return data_chunks



from .models.DataPoint import DataPoint
from .models.VectorConfig import VectorConfig
from .models.CollectionConfig import CollectionConfig
from .vector_db_interface import VectorDBInterface
from .config import get_vectordb_config
from .get_vector_engine import get_vector_engine
